/* 
   country.c

   THIS SOFTWARE IS IN PUBLIC DOMAIN
     There is no limitation on any kind of usage of this code. If you
     modify this code please remove this header and rename the file.
     
     You use it because it is yours.
     
     Originally written and maintained by Matlib (matlib@matlibhax.com),
     any comments or suggestions are welcome.
     
   Change log:
        1  2007-03-04  Matlib
        2  2007-03-12  Matlib  + link against libarguments
        3  2009-02-19  Matlib  + allow multiple records from command line
                               + minor improvements
        4  2011-11-23  Matlib  + link against libarguments 4
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <pthread.h>
#include <semaphore.h>

#include <arguments.h>


struct DB_Record 
{
	struct in_addr to;
	char country_code [2];
};

struct DB_Record *database;


#define ERROR_CHECK(_CONDITION_,_FUNCTION_) \
	if ( _CONDITION_ ) \
	{ \
		perror ( _FUNCTION_ ); \
		exit (EXIT_FAILURE); \
	}
#define ERROR_CHECKR(_RC_,_CALL_,_FUNCTION_) \
	{ \
		_RC_ = _CALL_; \
		if (_RC_ != 0) \
		{ \
			fprintf (stderr, "%s: %s\n", _FUNCTION_, strerror (_RC_)); \
			exit (EXIT_FAILURE); \
		}\
	}



static void address_to_num (char *address, struct in_addr *num)
{
	if (inet_aton (address, num) == 0)
	{
		fprintf (stderr, "Addres is invalid: \"%s\"\n", address);
		exit (EXIT_FAILURE);
	}
	num -> s_addr = ntohl (num -> s_addr);  
}

static int port_no (char *service)
{
	int port;
	struct servent *s;
	char *tail;

	port = strtol (service, &tail, 0);
	if ((port > 0xFFFF) || (*tail != 0))
	{
		s = getservbyname (service, "tcp");
		if (s == NULL)
		{
			fprintf (stderr, "getservbyname: No such service: %s\n", service);
			return EXIT_FAILURE;
		}
		return s -> s_port;
	}
	else
		return htons (port);
}






typedef int Store_Record (int, char *, char *, char *);

static int store_record_ignore (int i, char *from, char *to, char *country)
{
	return 1;
}


struct in_addr from_addr;

static int store_record (int i, char *from, char *to, char *country)
{
	address_to_num (from, &from_addr);
	if ((i > 0) && (from_addr.s_addr != database [i - 1].to.s_addr + 1))
		fprintf (stderr, "WARNING: gap detected @ %s\n", from);
	address_to_num (to, &(database [i].to));

	database [i].country_code [0] = country [0];
	database [i].country_code [1] = country [1];
	return 1;
}



int store_lookup (int i, char *from, char *to, char *country)
{
	static struct in_addr to_addr;
	address_to_num (to, &to_addr);
	if (from_addr.s_addr <= to_addr.s_addr)
	{
		puts (country);
		return 0;
	}
	return 1;
}



char *parse_record (char *string)
{
	int i;

	for (i = 0; (string [i] != ' ') && (string [i] != '\t'); i ++)
		if (string [i] == 0)
			return NULL;
	string [i] = 0;
	return string + i + 1;
}


int load_database (Store_Record *store_db_record, char *database_file)
{
	FILE *country_db;
	char line [100];
	char *from, *to, *country;
	int l, i;

	country_db = fopen (database_file, "r");
	ERROR_CHECK (country_db == NULL, database_file);
	for (i = 0; fgets (line, sizeof (line), country_db) != NULL; i ++)
	{
		l = strlen (line);
		if (l > 0)
		{
			line [l - 1] = 0;
			from = line;
			to = parse_record (from);
			country = parse_record (to);
			if ((to != NULL) && (country != NULL) && (strlen (from) > 6) 
				 && (strlen (to) > 6) && (strlen (country) > 0))
			if (store_db_record (i, from, to, 
				 country [0] == '-' ? "--" : country) == 0)
				break;
		}
	}
	ERROR_CHECK (fclose (country_db) != 0, "fclose");
	return i;
}





static int seek_address (char *address)
{
	int i;

	if (inet_aton (address, &from_addr) == 0)
		return -1;
	from_addr.s_addr = ntohl (from_addr.s_addr);
	for (i = 0; database [i].to.s_addr < from_addr.s_addr; i ++);
	return i;
}




struct Argument arguments [] =
{
  {'i', "interactive", argument_bool,   ARGUMENT_F_ONCE, {.value_u = 0}},
  {'b', "backlog",     argument_u,      ARGUMENT_F_PAR | ARGUMENT_F_ONCE, {.value_u = 5}},
  {'m', "max-conn",    argument_u,      ARGUMENT_F_PAR | ARGUMENT_F_ONCE, {.value_u = 20}},
  {'p', "port",        argument_string, ARGUMENT_F_PAR | ARGUMENT_F_ONCE, {.value_str = "2003"}},
  {'d', "database",    argument_string, ARGUMENT_F_PAR | ARGUMENT_F_ONCE, {.value_str = "/etc/country"}},
  {}
};
#define ARG_INTR ((arguments [0].value) != 0)
#define ARG_BLOG ((int) (arguments [1].value))
#define ARG_MAXC ((int) (arguments [2].value))
#define ARG_PORT ((char *) (arguments [3].value))
#define ARG_DBAS ((char *) (arguments [4].value))
#define ARG_F_BLOG (((arguments [1].flags) & ARGUMENT_F_OCC) != 0)
#define ARG_F_MAXC (((arguments [2].flags) & ARGUMENT_F_OCC) != 0)
#define ARG_F_PORT (((arguments [3].flags) & ARGUMENT_F_OCC) != 0)


sem_t ccnt_s;

static void process_connection (int socket)
{
	char buffer [32];
	int i;

loop:
	i = 0;
	buffer [0] = 0;
	do
	{
		if (buffer [i] != '\r')
			i ++;
		if ((i >= sizeof (buffer)) 
			 || (read (socket, buffer + i, 1) != 1))
		{
			close (socket);
			return;
		}
	}
	while (buffer [i] != '\n');
	buffer [i] = 0;
	if ((i >= 8) && (i <= 21))
	{
		i = seek_address (buffer + 1);
		if (i >= 0)
		{
			write (socket, &(database [i].country_code), 2);
			write (socket, "\n", 1);
		}
	}
	goto loop;
}

static void *process_connection_thread (void *socket)
{
	process_connection ((int) socket);
	sem_post (&ccnt_s);
	return NULL;
}

void block_signal (int sig)
{
	sigset_t set;

	sigemptyset (&set);
	ERROR_CHECK (sigaddset (&set, sig) != 0, "sigaddset");
	ERROR_CHECK (sigprocmask (SIG_BLOCK, &set, NULL) != 0, "sigprocmask");
}

void listen_for_connections ()
{
	int l, a, rc;
	size_t s;
	struct sockaddr_in name;
	pthread_t tid;
	pid_t pid;

	ERROR_CHECK (sem_init (&ccnt_s, 0, ARG_MAXC) != 0, "sem_init");
	l = socket (PF_INET, SOCK_STREAM, 0);
	ERROR_CHECK (l == -1, "socket");
	name.sin_family = AF_INET;
	name.sin_port = port_no (ARG_PORT);
	name.sin_addr.s_addr = htons (INADDR_ANY);
	ERROR_CHECK (bind (l, (void *) &name, sizeof (name)) == -1, "bind");
	ERROR_CHECK (listen (l, ARG_BLOG) == -1, "listen");
	pid = fork ();
	ERROR_CHECK (pid == -1, "fork");
	if (pid != 0)
		return;
	block_signal (SIGHUP);
	block_signal (SIGPIPE);
	close (STDIN_FILENO);
	close (STDOUT_FILENO);
loop:
	sem_wait (&ccnt_s);
	s = sizeof (name);
	a = accept (l, (void *) &name, &s);
	ERROR_CHECK (a == -1, "accept");
	ERROR_CHECKR (rc, pthread_create (&tid, NULL, 
		 process_connection_thread, (void *) a), "pthread_create");
	ERROR_CHECKR (rc, pthread_detach (tid), "pthread_detach")
	goto loop;
}




int main (int const argc, char const *const *const argv)
{
	int n;

	n = arguments_parse ("country", "4 (2011-11-23)", "[ip]...",
		 "IP to country database resolver/daemon", argc, argv, arguments);
	if (n < argc)
	{
		if (ARG_INTR)
			arguments_error_freeform ("Command line query does not work "
				 "in interactive mode.");
		if (ARG_F_PORT || ARG_F_MAXC || ARG_F_BLOG)
			arguments_error_freeform ("Command line query does not work "
				 "in daemon mode.");
		while (n < argc)
		{
			if (inet_aton (argv [n], &from_addr) == 0)
			{
				fprintf (stderr, "Address is invalid: \"%s\"\n", argv [n]);
				exit (EXIT_FAILURE);
			}
			from_addr.s_addr = ntohl (from_addr.s_addr);
			load_database (store_lookup, ARG_DBAS);
			n ++;
		}
		return EXIT_FAILURE;
	}

	from_addr.s_addr = -1;
	n = load_database (store_record_ignore, ARG_DBAS);
	database = malloc (n * sizeof (struct DB_Record));
	ERROR_CHECK (database == NULL, "malloc");
	load_database (store_record, ARG_DBAS);
	if (ARG_INTR)
	{
		if (ARG_F_PORT || ARG_F_MAXC || ARG_F_BLOG)
			arguments_error_freeform ("Standard input not work in daemon mode.");
		process_connection (STDIN_FILENO);
	}
	else
		listen_for_connections ();
	return EXIT_SUCCESS;
}

